# coding=utf-8
import shutil

from bscearth.utils.log import Log

from earthdiagnostics.diagnostic import Diagnostic, DiagnosticOption, DiagnosticDomainOption, DiagnosticListIntOption
from earthdiagnostics.frequency import Frequencies
from earthdiagnostics.utils import Utils, TempFile
from earthdiagnostics.variable_type import VariableType
from calendar import monthrange


class MonthlyPercentile(Diagnostic):
    """
    Calculates the montlhy percentiles

    :param data_manager: data management object
    :type data_manager: DataManager
    :param startdate: startdate
    :type startdate: str
    :param member: member number
    :type member: int
    :param chunk: chunk's number
    :type chunk: int
    :param variable: variable to average
    :type variable: str
    """

    alias = 'monpercent'
    "Diagnostic alias for the configuration file"

    def __init__(self, data_manager, startdate, member, chunk, domain, variable, percentiles):
        Diagnostic.__init__(self, data_manager)
        self.startdate = startdate
        self.member = member
        self.chunk = chunk
        self.variable = variable
        self.domain = domain
        self.percentiles = percentiles

    def __eq__(self, other):
        return self.startdate == other.startdate and self.member == other.member and self.chunk == other.chunk and \
            self.domain == other.domain and self.variable == other.variable and self.percentiles == other.percentiles

    def __str__(self):
        return 'Monthly percentile Startdate: {0} Member: {1} Chunk: {2} ' \
               'Variable: {3}:{4} Percentiles: {5}'.format(self.startdate, self.member, self.chunk,
                                                           self.domain, self.variable,
                                                           ', '.join(str(i) for i in self.percentiles))

    @classmethod
    def generate_jobs(cls, diags, options):
        """
        Creates a job for each chunk to compute the diagnostic

        :param diags: Diagnostics manager class
        :type diags: Diags
        :param options: domain, variable, percentil number, maximum depth (level)
        :type options: list[str]
        :return:
        """
        options_available = (DiagnosticOption('domain'),
                             DiagnosticDomainOption('variable'),
                             DiagnosticListIntOption('percentiles', None, 0, 100))
        options = cls.process_options(options, options_available)

        job_list = list()
        for startdate, member, chunk in diags.config.experiment.get_chunk_list():
            job_list.append(MonthlyPercentile(diags.data_manager, startdate, member, chunk,
                                              options['variable'], options['domain'], options['percentiles']))
        return job_list

    def compute(self):
        """
        Runs the diagnostic
        """
        variable_file = self.data_manager.get_file(self.domain, self.variable, self.startdate, self.member, self.chunk)
        temp = TempFile.get()
        handler = Utils.openCdf(variable_file)
        datetimes = Utils.get_datetime_from_netcdf(handler)
        handler.close()

        start_index = 0
        while datetimes[start_index].day != 1 and start_index < handler.size:
            start_index += 1
        if start_index == datetimes.size:
            raise Exception('No complete month for diagnostic {0}'.format(self))

        end_index = datetimes.size - 1
        while datetimes[end_index].day != monthrange(datetimes[end_index].year, datetimes[end_index].month)[1] \
                and end_index >= 0:
            end_index -= 1

        if end_index < 0:
            raise Exception('No complete month for diagnostic {0}'.format(self))

        if start_index != 0 or end_index != datetimes.size - 1:
            start_date = '{0.year}-{0.month}-{0.day}'.format(datetimes[start_index])
            end_date = '{0.year}-{0.month}-{0.day}'.format(datetimes[end_index])
            Utils.cdo.seldate('{0},{1}'.format(start_date, end_date), input=variable_file, output=temp)
            Utils.rename_variable(temp, 'lev', 'ensemble', False, True)
            shutil.move(temp, variable_file)

        Log.debug('Computing minimum')
        monmin_file = TempFile.get()
        Utils.cdo.monmin(input=variable_file, output=monmin_file)

        Log.debug('Computing maximum')
        monmax_file = TempFile.get()
        Utils.cdo.monmax(input=variable_file, output=monmax_file)

        for percentile in self.percentiles:
            Log.debug('Computing percentile {0}', percentile)
            Utils.cdo.monpctl(str(percentile), input=[variable_file, monmin_file, monmax_file], output=temp)
            Utils.rename_variable(temp, 'lev', 'ensemble', False, True)
            self.send_file(temp, self.domain, '{0}_q{1}'.format(self.variable, percentile), self.startdate,
                           self.member, self.chunk, frequency=Frequencies.monthly, rename_var=self.variable,
                           vartype=VariableType.STATISTIC)



